import pprint


def process_objects(all_objects):
    """Worker function handling incoming objects"""
    for (object_type, objects) in all_objects.items():
        for object_ in objects:
            print(f"New {object_type} object:")
            pprint.pprint(object_)
            print()


def main():
    from swh.journal.client import get_journal_client

    # Usually read from a config file:
    config = {
        "brokers": ["localhost:9092"],
        "group_id": "my-consumer-group",
        "auto_offset_reset": "earliest",
    }

    # Initialize the client
    client = get_journal_client(
        "kafka", object_types=["revision", "release"], privileged=True, **config
    )

    try:
        # Run the client forever
        client.process(process_objects)
    except KeyboardInterrupt:
        print("Called Ctrl-C, exiting.")
        exit(0)


if __name__ == "__main__":
    main()



# Copyright (C) 2017-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import defaultdict
import enum
from importlib import import_module
from itertools import cycle
import logging
import os
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
import warnings

from confluent_kafka import (
    OFFSET_BEGINNING,
    Consumer,
    KafkaError,
    KafkaException,
    TopicPartition,
)

from swh.core.statsd import Statsd
from swh.journal import DEFAULT_PREFIX

from .serializers import kafka_to_value

logger = logging.getLogger(__name__)
rdkafka_logger = logging.getLogger(__name__ + ".rdkafka")


# Only accepted offset reset policy accepted
ACCEPTED_OFFSET_RESET = ["earliest", "latest"]

# Errors that Kafka raises too often and are not useful; therefore they
# we lower their log level to DEBUG instead of INFO.
_SPAMMY_ERRORS = [
    KafkaError._NO_OFFSET,
]


class EofBehavior(enum.Enum):
    """Possible behaviors when reaching the end of the log"""

    CONTINUE = "continue"
    STOP = "stop"
    RESTART = "restart"


def get_journal_client(cls: str, **kwargs: Any):
    """Factory function to instantiate a journal client object.

    Currently, only the "kafka" journal client is supported.
    """
    if cls == "kafka":
        if "stats_cb" in kwargs:
            stats_cb = kwargs["stats_cb"]
            if isinstance(stats_cb, str):
                try:
                    module_path, func_name = stats_cb.split(":")
                except ValueError:
                    raise ValueError(
                        "Invalid stats_cb configuration option: "
                        "it should be a string like 'path.to.module:function'"
                    )
                try:
                    module = import_module(module_path, package=__package__)
                except ModuleNotFoundError:
                    raise ValueError(
                        "Invalid stats_cb configuration option: "
                        f"module {module_path} not found"
                    )
                try:
                    kwargs["stats_cb"] = getattr(module, func_name)
                except AttributeError:
                    raise ValueError(
                        "Invalid stats_cb configuration option: "
                        f"function {func_name} not found in module {module_path}"
                    )
        return JournalClient(**kwargs)
    raise ValueError("Unknown journal client class `%s`" % cls)


def _error_cb(error):
    if error.fatal():
        raise KafkaException(error)
    if error.code() in _SPAMMY_ERRORS:
        logger.debug("Received non-fatal kafka error: %s", error)
    else:
        logger.info("Received non-fatal kafka error: %s", error)


def _on_commit(error, partitions):
    if error is not None:
        _error_cb(error)


class JournalClient:
    """A base client for the Software Heritage journal.

    The current implementation of the journal uses Apache Kafka
    brokers to publish messages under a given topic prefix, with each
    object type using a specific topic under that prefix. If the `prefix`
    argument is None (default value), it will take the default value
    `'swh.journal.objects'`.

    Clients subscribe to events specific to each object type as listed in the
    ``object_types`` argument (if unset, defaults to all existing kafka topic under
    the prefix).

    Clients can be sharded by setting the ``group_id`` to a common
    value across instances. The journal will share the message
    throughput across the nodes sharing the same group_id.

    Messages are processed by the ``worker_fn`` callback passed to the `process`
    method, in batches of maximum ``batch_size`` messages (defaults to 200).

    The objects passed to the ``worker_fn`` callback are the result of the kafka
    message converted by the ``value_deserializer`` function. By default (if this
    argument is not given), it will produce dicts (using the ``kafka_to_value``
    function). This signature of the function is::

        value_deserializer(object_type: str, kafka_msg: bytes) -> Any

    If the value returned by ``value_deserializer`` is None, it is ignored and
    not passed the ``worker_fn`` function.

    Arguments:
        stop_after_objects: If set, the processing stops after processing
            this number of messages in total.
        on_eof: What to do when reaching the end of each partition (keep consuming,
            stop, or restart from earliest offsets); defaults to continuing.
            This can be either a :class:`EofBehavior` variant or a string containing the
            name of one of the variants.
        stop_on_eof: (deprecated) equivalent to passing ``on_eof=EofBehavior.STOP``
        auto_offset_reset: sets the behavior of the client when the consumer group
            initializes: ``'earliest'`` (the default) processes all objects since the
            inception of the topics; ``''``

    Any other named argument is passed directly to KafkaConsumer().

    """

    def __init__(
        self,
        brokers: Union[str, List[str]],
        group_id: str,
        prefix: Optional[str] = None,
        object_types: Optional[List[str]] = None,
        privileged: bool = False,
        stop_after_objects: Optional[int] = None,
        batch_size: int = 200,
        process_timeout: Optional[float] = None,
        auto_offset_reset: str = "earliest",
        stop_on_eof: Optional[bool] = None,
        on_eof: Optional[Union[EofBehavior, str]] = None,
        value_deserializer: Optional[Callable[[str, bytes], Any]] = None,
        **kwargs,
    ):
        if prefix is None:
            prefix = DEFAULT_PREFIX
        if auto_offset_reset not in ACCEPTED_OFFSET_RESET:
            raise ValueError(
                "Option 'auto_offset_reset' only accept %s, not %s"
                % (ACCEPTED_OFFSET_RESET, auto_offset_reset)
            )

        if batch_size <= 0:
            raise ValueError("Option 'batch_size' needs to be positive")
        if value_deserializer:
            self.value_deserializer = value_deserializer
        else:
            self.value_deserializer = lambda _, value: kafka_to_value(value)
        if stop_on_eof is not None:
            if on_eof is not None:
                raise TypeError(
                    "stop_on_eof and on_eof are mutually exclusive (the former is "
                    "deprecated)"
                )
            elif stop_on_eof:
                warnings.warn(
                    "stop_on_eof=True should be replaced with "
                    "on_eof=EofBehavior.STOP ('on_eof: stop' in YAML)",
                    DeprecationWarning,
                    2,
                )
                on_eof = EofBehavior.STOP
            else:
                warnings.warn(
                    "stop_on_eof=False should be replaced with "
                    "on_eof=EofBehavior.CONTINUE ('on_eof: continue' in YAML)",
                    DeprecationWarning,
                    2,
                )
                on_eof = EofBehavior.CONTINUE
        self.on_eof = EofBehavior(on_eof or EofBehavior.CONTINUE)

        if isinstance(brokers, str):
            brokers = [brokers]

        debug_logging = rdkafka_logger.isEnabledFor(logging.DEBUG)
        if debug_logging and "debug" not in kwargs:
            kwargs["debug"] = "consumer"

        # Static group instance id management
        group_instance_id = os.environ.get("KAFKA_GROUP_INSTANCE_ID")
        if group_instance_id:
            kwargs["group.instance.id"] = group_instance_id

        if "group.instance.id" in kwargs:
            # When doing static consumer group membership, set a higher default
            # session timeout. The session timeout is the duration after which
            # the broker considers that a consumer has left the consumer group
            # for good, and triggers a rebalance. Considering our current
            # processing pattern, 10 minutes gives the consumer ample time to
            # restart before that happens.
            if "session.timeout.ms" not in kwargs:
                kwargs["session.timeout.ms"] = 10 * 60 * 1000  # 10 minutes

        if "session.timeout.ms" in kwargs:
            # When the session timeout is set, rdkafka requires the max poll
            # interval to be set to a higher value; the max poll interval is
            # rdkafka's way of figuring out whether the client's message
            # processing thread has stalled: when the max poll interval lapses
            # between two calls to consumer.poll(), rdkafka leaves the consumer
            # group and terminates the connection to the brokers.
            #
            # We default to 1.5 times the session timeout
            if "max.poll.interval.ms" not in kwargs:
                kwargs["max.poll.interval.ms"] = kwargs["session.timeout.ms"] // 2 * 3

        consumer_settings = {
            **kwargs,
            "bootstrap.servers": ",".join(brokers),
            "auto.offset.reset": auto_offset_reset,
            "group.id": group_id,
            "on_commit": _on_commit,
            "error_cb": _error_cb,
            "enable.auto.commit": False,
            "logger": rdkafka_logger,
        }

        if self.on_eof != EofBehavior.CONTINUE:
            consumer_settings["enable.partition.eof"] = True

        if logger.isEnabledFor(logging.DEBUG):
            filtered_keys = {"sasl.password"}
            logger.debug("Consumer settings:")
            for k, v in consumer_settings.items():
                if k in filtered_keys:
                    v = "**filtered**"

                logger.debug("    %s: %s", k, v)

        self.statsd = Statsd(
            namespace="swh_journal_client", constant_tags={"group": group_id}
        )

        self.consumer = Consumer(consumer_settings)
        if privileged:
            privileged_prefix = f"{prefix}_privileged"
        else:  # do not attempt to subscribe to privileged topics
            privileged_prefix = f"{prefix}"
        existing_topics = [
            topic
            for topic in self.consumer.list_topics(timeout=10).topics.keys()
            if (
                topic.startswith(f"{prefix}.")
                or topic.startswith(f"{privileged_prefix}.")
            )
        ]
        if not existing_topics:
            raise ValueError(
                f"The prefix {prefix} does not match any existing topic "
                "on the kafka broker"
            )

        if not object_types:
            object_types = list({topic.split(".")[-1] for topic in existing_topics})

        self.subscription = []
        unknown_types = []
        for object_type in object_types:
            topics = (f"{privileged_prefix}.{object_type}", f"{prefix}.{object_type}")
            for topic in topics:
                if topic in existing_topics:
                    self.subscription.append(topic)
                    break
            else:
                unknown_types.append(object_type)
        if unknown_types:
            raise ValueError(
                f"Topic(s) for object types {','.join(unknown_types)} "
                "are unknown on the kafka broker"
            )

        logger.debug(f"Upstream topics: {existing_topics}")
        self.subscribe()

        self.stop_after_objects = stop_after_objects

        self.eof_reached: Set[Tuple[str, str]] = set()
        self.batch_size = batch_size

        if process_timeout is not None:
            raise DeprecationWarning(
                "'process_timeout' argument is not supported anymore by "
                "JournalClient; please remove it from your configuration.",
            )

    def subscribe(self):
        """Subscribe to topics listed in self.subscription

        This can be overridden if you need, for instance, to manually assign partitions.
        """
        logger.debug(f"Subscribing to: {self.subscription}")
        self.consumer.subscribe(topics=self.subscription)

    def process(self, worker_fn: Callable[[Dict[str, List[dict]]], None]):
        """Polls Kafka for a batch of messages, and calls the worker_fn
        with these messages.

        Args:
            worker_fn: Function called with the messages as argument.
        """
        total_objects_processed = 0
        # timeout for message poll
        timeout = 1.0

        with self.statsd.status_gauge(
            "status", statuses=["idle", "processing", "waiting"]
        ) as set_status:
            set_status("idle")
            while True:
                batch_size = self.batch_size
                if self.stop_after_objects:
                    if total_objects_processed >= self.stop_after_objects:
                        break

                    # clamp batch size to avoid overrunning stop_after_objects
                    batch_size = min(
                        self.stop_after_objects - total_objects_processed,
                        batch_size,
                    )
                set_status("waiting")
                for i in cycle(reversed(range(10))):
                    messages = self.consumer.consume(
                        timeout=timeout, num_messages=batch_size
                    )
                    if messages:
                        break

                    # do check for an EOF condition iff we already consumed
                    # messages, otherwise we could detect an EOF condition
                    # before messages had a chance to reach us (e.g. in tests)
                    if total_objects_processed > 0 and i == 0:
                        if self.on_eof == EofBehavior.STOP:
                            at_eof = all(
                                (tp.topic, tp.partition) in self.eof_reached
                                for tp in self.consumer.assignment()
                            )
                            if at_eof:
                                break
                        elif self.on_eof == EofBehavior.RESTART:
                            for tp in self.consumer.assignment():
                                if (tp.topic, tp.partition) in self.eof_reached:
                                    self.eof_reached.remove((tp.topic, tp.partition))
                                    self.statsd.increment("partition_restart_total")
                                    new_tp = TopicPartition(
                                        tp.topic,
                                        tp.partition,
                                        OFFSET_BEGINNING,
                                    )
                                    self.consumer.seek(new_tp)
                        elif self.on_eof == EofBehavior.CONTINUE:
                            pass  # Nothing to do, we'll just keep consuming
                        else:
                            assert False, f"Unexpected on_eof behavior: {self.on_eof}"

                if messages:
                    set_status("processing")
                    batch_processed, at_eof = self.handle_messages(messages, worker_fn)

                    set_status("idle")
                    # report the number of handled messages
                    self.statsd.increment("handle_message_total", value=batch_processed)
                    total_objects_processed += batch_processed

                if self.on_eof == EofBehavior.STOP and at_eof:
                    self.statsd.increment("stop_total")
                    break

        return total_objects_processed

    def handle_messages(
        self, messages, worker_fn: Callable[[Dict[str, List[dict]]], None]
    ) -> Tuple[int, bool]:
        objects: Dict[str, List[Any]] = defaultdict(list)
        nb_processed = 0

        for message in messages:
            error = message.error()
            if error is not None:
                if error.code() == KafkaError._PARTITION_EOF:
                    self.eof_reached.add((message.topic(), message.partition()))
                else:
                    _error_cb(error)
                continue
            if message.value() is None:
                # ignore message with no payload, these can be generated in tests
                continue
            nb_processed += 1
            object_type = message.topic().split(".")[-1]
            deserialized_object = self.deserialize_message(
                message, object_type=object_type
            )
            if deserialized_object is not None:
                objects[object_type].append(deserialized_object)

        if objects:
            worker_fn(dict(objects))
        self.consumer.commit()

        if self.on_eof in (EofBehavior.STOP, EofBehavior.RESTART):
            at_eof = all(
                (tp.topic, tp.partition) in self.eof_reached
                for tp in self.consumer.assignment()
            )
        elif self.on_eof == EofBehavior.CONTINUE:
            at_eof = False
        else:
            assert False, f"Unexpected on_eof behavior: {self.on_eof}"

        return nb_processed, at_eof

    def deserialize_message(self, message, object_type=None):
        return self.value_deserializer(object_type, message.value())

    def close(self):
        self.consumer.close()



# Copyright (C) 2019-2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from multiprocessing import Manager
from typing import Any, Callable, Dict, Iterable, List, Tuple

from .interface import ValueProtocol

logger = logging.getLogger(__name__)


class InMemoryJournalWriter:
    objects: List[Tuple[str, ValueProtocol]]
    privileged_objects: List[Tuple[str, ValueProtocol]]

    def __init__(
        self,
        value_sanitizer: Callable[[str, Dict[str, Any]], Dict[str, Any]],
        anonymize: bool = False,
    ):
        # Share the list of objects across processes, for RemoteAPI tests.
        self.manager = Manager()
        self.objects = self.manager.list()  # type: ignore[assignment]
        self.privileged_objects = self.manager.list()  # type: ignore[assignment]
        self.anonymize = anonymize

    def write_addition(self, object_type: str, object_: ValueProtocol) -> None:
        object_.unique_key()  # Check this does not error, to mimic the kafka writer
        anon_object_ = None
        if self.anonymize:
            anon_object_ = object_.anonymize()
        if anon_object_ is not None:
            self.privileged_objects.append((object_type, object_))
            self.objects.append((object_type, anon_object_))
        else:
            self.objects.append((object_type, object_))

    def write_additions(
        self, object_type: str, objects: Iterable[ValueProtocol]
    ) -> None:
        for object_ in objects:
            self.write_addition(object_type, object_)

    def flush(self) -> None:
        pass



# Copyright (C) 2019-2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.journal.writer import model_object_dict_sanitizer
from swh.journal.writer.inmemory import InMemoryJournalWriter
from swh.model.model import BaseModel
from swh.model.tests.swh_model_data import TEST_OBJECTS


def test_write_additions_anonymized():
    writer = InMemoryJournalWriter(
        value_sanitizer=model_object_dict_sanitizer, anonymize=True
    )
    expected = []
    priv_expected = []

    for object_type, objects in TEST_OBJECTS.items():
        writer.write_additions(object_type, objects)

        for object in objects:
            if object.anonymize():
                expected.append((object_type, object.anonymize()))
                priv_expected.append((object_type, object))
            else:
                expected.append((object_type, object))

    assert set(priv_expected) == set(writer.privileged_objects)
    assert set(expected) == set(writer.objects)


def test_write_additions():
    writer = InMemoryJournalWriter(value_sanitizer=model_object_dict_sanitizer)
    expected = set()

    for object_type, objects in TEST_OBJECTS.items():
        writer.write_additions(object_type, objects)

        for object in objects:
            expected.add((object_type, object))

    assert not set(writer.privileged_objects)
    assert expected == set(writer.objects)


def test_write_addition_errors_without_unique_key():
    writer = InMemoryJournalWriter(value_sanitizer=model_object_dict_sanitizer)

    with pytest.raises(NotImplementedError):
        writer.write_addition("BaseModel", BaseModel())



# Copyright (C) 2021-2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import io
from typing import Dict, List, Tuple

from swh.journal.serializers import kafka_stream_to_value
from swh.journal.writer import get_journal_writer, model_object_dict_sanitizer
from swh.journal.writer.interface import JournalWriterInterface
from swh.model.tests.swh_model_data import TEST_OBJECTS


def fill_writer(writer: JournalWriterInterface) -> List[Tuple[str, Dict]]:
    expected = []
    for object_type, objects in TEST_OBJECTS.items():
        writer.write_additions(object_type, objects)

        for object in objects:
            objd = object.to_dict()
            if object_type == "content":
                objd.pop("data")

            expected.append((object_type, objd))
    writer.flush()
    return expected


def test_stream_journal_writer_stream():
    outs = io.BytesIO()

    writer = get_journal_writer(
        cls="stream",
        value_sanitizer=model_object_dict_sanitizer,
        output_stream=outs,
    )
    expected = fill_writer(writer)

    outs.seek(0, 0)
    unpacker = kafka_stream_to_value(outs)
    for i, (objtype, objd) in enumerate(unpacker, start=1):
        assert (objtype, objd) in expected
    assert len(expected) == i


def test_stream_journal_writer_filename(tmp_path):
    out_fname = str(tmp_path / "journal.msgpack")

    writer = get_journal_writer(
        cls="stream",
        value_sanitizer=model_object_dict_sanitizer,
        output_stream=out_fname,
    )
    expected = fill_writer(writer)

    with open(out_fname, "rb") as outs:
        unpacker = kafka_stream_to_value(outs)
        for i, (objtype, objd) in enumerate(unpacker, start=1):
            assert (objtype, objd) in expected
        assert len(expected) == i


def test_stream_journal_writer_stdout(capfdbinary):
    writer = get_journal_writer(
        cls="stream",
        value_sanitizer=model_object_dict_sanitizer,
        output_stream="-",
    )
    expected = fill_writer(writer)

    captured = capfdbinary.readouterr()
    assert captured.err == b""
    outs = io.BytesIO(captured.out)

    unpacker = kafka_stream_to_value(outs)
    for i, (objtype, objd) in enumerate(unpacker, start=1):
        assert (objtype, objd) in expected
    assert len(expected) == i



# Copyright (C) 2019-2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# deprecated, for BW compat only

import warnings

from swh.model.tests.swh_model_data import DUPLICATE_CONTENTS, TEST_OBJECTS  # noqa

warnings.warn(
    "This module is deprecated, please use swh.model.tests.swh_model_data instead",
    category=DeprecationWarning,
)



# Copyright (C) 2017-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import OrderedDict
from datetime import datetime, timedelta, timezone
import itertools
from typing import List

import pytest

from swh.journal import serializers
from swh.model.tests.swh_model_data import TEST_OBJECTS


def test_key_to_kafka_repeatable():
    """Check the kafka key encoding is repeatable"""
    base_dict = {
        "a": "foo",
        "b": "bar",
        "c": "baz",
    }

    key = serializers.key_to_kafka(base_dict)

    for dict_keys in itertools.permutations(base_dict):
        d = OrderedDict()
        for k in dict_keys:
            d[k] = base_dict[k]

        assert key == serializers.key_to_kafka(d)


def test_pprint_key():
    """Test whether get_key works on all our objects"""
    for object_type, objects in TEST_OBJECTS.items():
        for obj in objects:
            key = obj.unique_key()
            pprinted_key = serializers.pprint_key(key)
            assert isinstance(pprinted_key, str)

            if isinstance(key, dict):
                assert pprinted_key[0], pprinted_key[-1] == "{}"
                for dict_key in key.keys():
                    assert f"{dict_key}:" in pprinted_key

            if isinstance(key, bytes):
                assert pprinted_key == key.hex()


def test_kafka_to_key() -> None:
    """Standard back and forth serialization with keys"""
    # All KeyType(s)
    keys: List[serializers.KeyType] = [
        {
            "a": "foo",
            "b": "bar",
            "c": "baz",
        },
        {
            "a": b"foobarbaz",
        },
        b"foo",
    ]
    for object_type, objects in TEST_OBJECTS.items():
        for obj in objects:
            key = obj.unique_key()
            keys.append(key)

    for key in keys:
        ktk = serializers.key_to_kafka(key)
        v = serializers.kafka_to_key(ktk)

        assert v == key


# limits of supported int values by msgpack
MININT = -(2**63)
MAXINT = 2**64 - 1

intvalues = [
    MININT * 2,
    MININT - 1,
    MININT,
    MININT + 1,
    -10,
    0,
    10,
    MAXINT - 1,
    MAXINT,
    MAXINT + 1,
    MAXINT * 2,
]


@pytest.mark.parametrize("value", intvalues)
def test_encode_int(value):
    assert serializers.kafka_to_value(serializers.value_to_kafka(value)) == value


datevalues = [
    datetime.now(tz=timezone.utc),
    datetime.now(tz=timezone(timedelta(hours=-23, minutes=-59))),
    datetime.now(tz=timezone(timedelta(hours=23, minutes=59))),
    datetime(1, 1, 1, 1, 1, tzinfo=timezone.utc),
    datetime(2100, 1, 1, 1, 1, tzinfo=timezone.utc),
]


@pytest.mark.parametrize("value", datevalues)
def test_encode_datetime(value):
    assert serializers.kafka_to_value(serializers.value_to_kafka(value)) == value


@pytest.mark.parametrize("value", datevalues)
def test_encode_datetime_bw(value):
    bwdate = {b"swhtype": "datetime", b"d": value.isoformat()}
    assert serializers.kafka_to_value(serializers.value_to_kafka(bwdate)) == value



# Copyright (C) 2021-2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from typing import Any, BinaryIO, Callable, Dict, Iterable

from swh.journal.serializers import value_to_kafka

from .interface import ValueProtocol

logger = logging.getLogger(__name__)


class StreamJournalWriter:
    """A simple JournalWriter which serializes objects in a stream

    Might be used to serialize a storage in a file to generate a test dataset.
    """

    def __init__(
        self,
        output_stream: BinaryIO,
        value_sanitizer: Callable[[str, Dict[str, Any]], Dict[str, Any]],
    ):
        # Share the list of objects across processes, for RemoteAPI tests.
        self.output = output_stream
        self.value_sanitizer = value_sanitizer

    def write_addition(self, object_type: str, object_: ValueProtocol) -> None:
        object_.unique_key()  # Check this does not error, to mimic the kafka writer
        dict_ = self.value_sanitizer(object_type, object_.to_dict())
        self.output.write(value_to_kafka((object_type, dict_)))

    def write_additions(
        self, object_type: str, objects: Iterable[ValueProtocol]
    ) -> None:
        for object_ in objects:
            self.write_addition(object_type, object_)

    def flush(self) -> None:
        self.output.flush()



# Copyright (C) 2019-2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import sys
from typing import Any, BinaryIO, Dict, Type
import warnings

from .interface import JournalWriterInterface


def model_object_dict_sanitizer(
    object_type: str, object_dict: Dict[str, Any]
) -> Dict[str, str]:
    object_dict = object_dict.copy()
    if object_type == "content":
        object_dict.pop("data", None)
    return object_dict


def get_journal_writer(cls, **kwargs) -> JournalWriterInterface:

    if "args" in kwargs:
        warnings.warn(
            'Explicit "args" key is deprecated, use keys directly instead.',
            DeprecationWarning,
        )
        kwargs = kwargs["args"]

    kwargs.setdefault("value_sanitizer", model_object_dict_sanitizer)

    if cls == "inmemory":  # FIXME: Remove inmemory in due time
        warnings.warn(
            "cls = 'inmemory' is deprecated, use 'memory' instead", DeprecationWarning
        )
        cls = "memory"

    JournalWriter: Type[JournalWriterInterface]
    if cls == "memory":
        from .inmemory import InMemoryJournalWriter

        JournalWriter = InMemoryJournalWriter
    elif cls == "kafka":
        from .kafka import KafkaJournalWriter

        JournalWriter = KafkaJournalWriter
    elif cls == "stream":
        from .stream import StreamJournalWriter

        JournalWriter = StreamJournalWriter

        assert "output_stream" in kwargs
        outstream: BinaryIO
        if kwargs["output_stream"] in ("-", b"-"):
            outstream = os.fdopen(sys.stdout.fileno(), "wb", closefd=False)
        elif isinstance(kwargs["output_stream"], (str, bytes)):
            outstream = open(kwargs["output_stream"], "wb")
        else:
            outstream = kwargs["output_stream"]
        kwargs["output_stream"] = outstream
    else:
        raise ValueError("Unknown journal writer class `%s`" % cls)

    return JournalWriter(**kwargs)



# Copyright (C) 2018-2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from typing import Iterable

from confluent_kafka import Consumer, Producer
import pytest

from swh.journal.pytest_plugin import assert_all_objects_consumed, consume_messages
from swh.journal.writer import model_object_dict_sanitizer
from swh.journal.writer.kafka import KafkaDeliveryError, KafkaJournalWriter
from swh.model.model import BaseModel, Directory, Release, Revision
from swh.model.tests.swh_model_data import TEST_OBJECTS


def test_kafka_writer(
    kafka_prefix: str,
    kafka_server: str,
    consumer: Consumer,
    privileged_object_types: Iterable[str],
):
    writer = KafkaJournalWriter(
        brokers=[kafka_server],
        client_id="kafka_writer",
        prefix=kafka_prefix,
        value_sanitizer=model_object_dict_sanitizer,
        anonymize=False,
    )

    expected_messages = 0

    for object_type, objects in TEST_OBJECTS.items():
        writer.write_additions(object_type, objects)
        expected_messages += len(objects)

    consumed_messages = consume_messages(consumer, kafka_prefix, expected_messages)
    assert_all_objects_consumed(consumed_messages)

    for key, obj_dict in consumed_messages["revision"]:
        obj = Revision.from_dict(obj_dict)
        for person in (obj.author, obj.committer):
            assert not (
                len(person.fullname) == 32
                and person.name is None
                and person.email is None
            )
    for key, obj_dict in consumed_messages["release"]:
        obj = Release.from_dict(obj_dict)
        # author is optional for release
        if obj.author is None:
            continue
        for person in (obj.author,):
            assert not (
                len(person.fullname) == 32
                and person.name is None
                and person.email is None
            )


def test_kafka_writer_anonymized(
    kafka_prefix: str,
    kafka_server: str,
    consumer: Consumer,
    privileged_object_types: Iterable[str],
):
    writer = KafkaJournalWriter(
        brokers=[kafka_server],
        client_id="kafka_writer",
        prefix=kafka_prefix,
        value_sanitizer=model_object_dict_sanitizer,
        anonymize=True,
    )

    expected_messages = 0

    for object_type, objects in TEST_OBJECTS.items():
        writer.write_additions(object_type, objects)
        expected_messages += len(objects)
        if object_type in privileged_object_types:
            expected_messages += len(objects)

    consumed_messages = consume_messages(consumer, kafka_prefix, expected_messages)
    assert_all_objects_consumed(consumed_messages, exclude=["revision", "release"])

    for key, obj_dict in consumed_messages["revision"]:
        obj = Revision.from_dict(obj_dict)
        for person in (obj.author, obj.committer):
            assert (
                len(person.fullname) == 32
                and person.name is None
                and person.email is None
            )
    for key, obj_dict in consumed_messages["release"]:
        obj = Release.from_dict(obj_dict)
        # author is optional for release
        if obj.author is None:
            continue
        for person in (obj.author,):
            assert (
                len(person.fullname) == 32
                and person.name is None
                and person.email is None
            )


def test_write_delivery_failure(kafka_prefix: str, kafka_server: str):
    class MockKafkaError:
        """A mocked kafka error"""

        def str(self):
            return "Mocked Kafka Error"

        def name(self):
            return "SWH_MOCK_ERROR"

    class KafkaJournalWriterFailDelivery(KafkaJournalWriter):
        """A journal writer which always fails delivering messages"""

        def _on_delivery(self, error, message):
            """Replace the inbound error with a fake delivery error"""
            super()._on_delivery(MockKafkaError(), message)

    kafka_prefix += ".swh.journal.objects"
    writer = KafkaJournalWriterFailDelivery(
        brokers=[kafka_server],
        client_id="kafka_writer",
        prefix=kafka_prefix,
        value_sanitizer=model_object_dict_sanitizer,
    )

    empty_dir = Directory(entries=())
    with pytest.raises(KafkaDeliveryError) as exc:
        writer.write_addition("directory", empty_dir)

    assert "Failed deliveries" in exc.value.message
    assert len(exc.value.delivery_failures) == 1
    delivery_failure = exc.value.delivery_failures[0]
    assert delivery_failure.key == empty_dir.id
    assert delivery_failure.code == "SWH_MOCK_ERROR"


def test_write_delivery_timeout(kafka_prefix: str, kafka_server: str):

    produced = []

    class MockProducer(Producer):
        """A kafka producer which pretends to produce messages, but never sends any
        delivery acknowledgements"""

        def produce(self, **kwargs):
            produced.append(kwargs)

    writer = KafkaJournalWriter(
        brokers=[kafka_server],
        client_id="kafka_writer",
        prefix=kafka_prefix,
        value_sanitizer=model_object_dict_sanitizer,
        flush_timeout=1,
        producer_class=MockProducer,
    )

    empty_dir = Directory(entries=())
    with pytest.raises(KafkaDeliveryError) as exc:
        writer.write_addition("directory", empty_dir)

    assert len(produced) == 1

    assert "timeout" in exc.value.message
    assert len(exc.value.delivery_failures) == 1
    delivery_failure = exc.value.delivery_failures[0]
    assert delivery_failure.key == empty_dir.id
    assert delivery_failure.code == "SWH_FLUSH_TIMEOUT"


class MockBufferErrorProducer(Producer):
    """A Kafka producer that returns a BufferError on the `n_buffererrors`
    first calls to produce."""

    def __init__(self, *args, **kwargs):
        self.n_buffererrors = kwargs.pop("n_bufferrors", 0)
        self.produce_calls = 0

        super().__init__(*args, **kwargs)

    def produce(self, **kwargs):
        self.produce_calls += 1
        if self.produce_calls <= self.n_buffererrors:
            raise BufferError("Local: Queue full")

        self.produce_calls = 0
        return super().produce(**kwargs)


def test_write_BufferError_retry(kafka_prefix: str, kafka_server: str, caplog):
    writer = KafkaJournalWriter(
        brokers=[kafka_server],
        client_id="kafka_writer",
        prefix=kafka_prefix,
        value_sanitizer=model_object_dict_sanitizer,
        flush_timeout=1,
        producer_class=MockBufferErrorProducer,
    )

    writer.producer.n_buffererrors = 4

    empty_dir = Directory(entries=())

    caplog.set_level(logging.DEBUG, "swh.journal.writer.kafka")
    writer.write_addition("directory", empty_dir)
    records = []
    for record in caplog.records:
        if "BufferError" in record.getMessage():
            records.append(record)

    assert len(records) == writer.producer.n_buffererrors


def test_write_BufferError_give_up(kafka_prefix: str, kafka_server: str, caplog):
    writer = KafkaJournalWriter(
        brokers=[kafka_server],
        client_id="kafka_writer",
        prefix=kafka_prefix,
        value_sanitizer=model_object_dict_sanitizer,
        flush_timeout=1,
        producer_class=MockBufferErrorProducer,
    )

    writer.producer.n_buffererrors = 5

    empty_dir = Directory(entries=())

    with pytest.raises(KafkaDeliveryError):
        writer.write_addition("directory", empty_dir)


def test_write_addition_errors_without_unique_key(kafka_prefix: str, kafka_server: str):
    writer = KafkaJournalWriter(
        brokers=[kafka_server],
        client_id="kafka_writer",
        prefix=kafka_prefix,
        value_sanitizer=model_object_dict_sanitizer,
    )

    with pytest.raises(NotImplementedError):
        writer.write_addition("BaseModel", BaseModel())



# Copyright (C) 2019-2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import time
from typing import Any, Callable, Dict, Iterable, List, NamedTuple, Optional, Type

from confluent_kafka import KafkaException, Producer

from swh.journal.serializers import KeyType, key_to_kafka, pprint_key, value_to_kafka

from .interface import ValueProtocol

logger = logging.getLogger(__name__)


class DeliveryTag(NamedTuple):
    """Unique tag allowing us to check for a message delivery"""

    topic: str
    kafka_key: bytes


class DeliveryFailureInfo(NamedTuple):
    """Verbose information for failed deliveries"""

    object_type: str
    key: KeyType
    message: str
    code: str


def get_object_type(topic: str) -> str:
    """Get the object type from a topic string"""
    return topic.rsplit(".", 1)[-1]


class KafkaDeliveryError(Exception):
    """Delivery failed on some kafka messages."""

    def __init__(self, message: str, delivery_failures: Iterable[DeliveryFailureInfo]):
        self.message = message
        self.delivery_failures = list(delivery_failures)

    def pretty_failures(self) -> str:
        return ", ".join(
            f"{f.object_type} {pprint_key(f.key)} ({f.message})"
            for f in self.delivery_failures
        )

    def __str__(self):
        return f"KafkaDeliveryError({self.message}, [{self.pretty_failures()}])"


class KafkaJournalWriter:
    """This class is used to write serialized versions of value objects to a series
    of Kafka topics. The type parameter of value objects, which must implement
    the `ValueProtocol`, is the type of values this writer will write.
    Typically, `ValueProtocol` will be `swh.model.model.BaseModel`.

    Topics used to send objects representations are built from a ``prefix`` plus the
    type of the object:

      ``{prefix}.{object_type}``

    Objects can be sent as is, or can be anonymized. The anonymization feature, when
    activated, will write anonymized versions of value objects in the main topic, and
    stock (non-anonymized) objects will be sent to a dedicated (privileged) set of
    topics:

      ``{prefix}_privileged.{object_type}``

    The anonymization of a value object is the result of calling its
    ``anonymize()`` method. An object is considered anonymizable if this
    method returns a (non-None) value.

    Args:
      brokers: list of broker addresses and ports.
      prefix: the prefix used to build the topic names for objects.
      client_id: the id of the writer sent to kafka.
      value_sanitizer: a function that takes the object type and the dict
        representation of an object as argument, and returns an other dict
        that should be actually stored in the journal (eg. removing keys
        that do no belong there)
      producer_config: extra configuration keys passed to the `Producer`.
      flush_timeout: timeout, in seconds, after which the `flush` operation
        will fail if some message deliveries are still pending.
      producer_class: override for the kafka producer class.
      anonymize: if True, activate the anonymization feature.
      auto_flush: if True (default), flush the kafka producer in
        ``write_addition()`` and ``write_additions()``. This should be set
        to False ONLY for testing purpuse. DO NOT USE ON PRODUCTION ENVIRONMENT.

    """

    def __init__(
        self,
        brokers: Iterable[str],
        prefix: str,
        client_id: str,
        value_sanitizer: Callable[[str, Dict[str, Any]], Dict[str, Any]],
        producer_config: Optional[Dict] = None,
        flush_timeout: float = 120,
        producer_class: Type[Producer] = Producer,
        anonymize: bool = False,
        auto_flush: bool = True,
    ):
        self._prefix = prefix
        self._prefix_privileged = f"{self._prefix}_privileged"
        self.anonymize = anonymize
        self.auto_flush = auto_flush

        if not producer_config:
            producer_config = {}

        if "message.max.bytes" not in producer_config:
            producer_config = {
                "message.max.bytes": 100 * 1024 * 1024,
                **producer_config,
            }

        self.producer = producer_class(
            {
                "bootstrap.servers": ",".join(brokers),
                "client.id": client_id,
                "on_delivery": self._on_delivery,
                "error_cb": self._error_cb,
                "logger": logger,
                "acks": "all",
                **producer_config,
            }
        )

        # Delivery management
        self.flush_timeout = flush_timeout

        # delivery tag -> original object "key" mapping
        self.deliveries_pending: Dict[DeliveryTag, KeyType] = {}

        # List of (object_type, key, error_msg, error_name) for failed deliveries
        self.delivery_failures: List[DeliveryFailureInfo] = []

        self.value_sanitizer = value_sanitizer

    def _error_cb(self, error):
        if error.fatal():
            raise KafkaException(error)
        logger.info("Received non-fatal kafka error: %s", error)

    def _on_delivery(self, error, message):
        (topic, key) = delivery_tag = DeliveryTag(message.topic(), message.key())
        sent_key = self.deliveries_pending.pop(delivery_tag, None)

        if error is not None:
            self.delivery_failures.append(
                DeliveryFailureInfo(
                    get_object_type(topic), sent_key, error.str(), error.name()
                )
            )

    def send(self, topic: str, key: KeyType, value):
        kafka_key = key_to_kafka(key)
        max_attempts = 5
        last_exception: Optional[Exception] = None
        for attempt in range(max_attempts):
            try:
                self.producer.produce(
                    topic=topic,
                    key=kafka_key,
                    value=value_to_kafka(value),
                )
            except BufferError as e:
                last_exception = e
                wait = 1 + 3 * attempt

                if logger.isEnabledFor(logging.DEBUG):  # pprint_key is expensive
                    logger.debug(
                        "BufferError producing %s %s; waiting for %ss",
                        get_object_type(topic),
                        pprint_key(kafka_key),
                        wait,
                    )
                self.producer.poll(wait)
            else:
                self.deliveries_pending[DeliveryTag(topic, kafka_key)] = key
                return

        # We reach this point if all delivery attempts have failed
        self.delivery_failures.append(
            DeliveryFailureInfo(
                get_object_type(topic), key, str(last_exception), "SWH_BUFFER_ERROR"
            )
        )

    def delivery_error(self, message) -> KafkaDeliveryError:
        """Get all failed deliveries, and clear them"""
        ret = self.delivery_failures
        self.delivery_failures = []

        while self.deliveries_pending:
            delivery_tag, orig_key = self.deliveries_pending.popitem()
            (topic, kafka_key) = delivery_tag
            ret.append(
                DeliveryFailureInfo(
                    get_object_type(topic),
                    orig_key,
                    "No delivery before flush() timeout",
                    "SWH_FLUSH_TIMEOUT",
                )
            )

        return KafkaDeliveryError(message, ret)

    def flush(self) -> None:
        start = time.monotonic()

        self.producer.flush(self.flush_timeout)

        while self.deliveries_pending:
            if time.monotonic() - start > self.flush_timeout:
                break
            self.producer.poll(0.1)

        if self.deliveries_pending:
            # Delivery timeout
            raise self.delivery_error(
                "flush() exceeded timeout (%ss)" % self.flush_timeout,
            )
        elif self.delivery_failures:
            raise self.delivery_error("Failed deliveries after flush()")

    def _write_addition(self, object_type: str, object_: ValueProtocol) -> None:
        """Write a single object to the journal"""
        key = object_.unique_key()

        if self.anonymize:
            anon_object_ = object_.anonymize()
            if anon_object_:  # can be either None, or an anonymized object
                # if the object is anonymizable, send the non-anonymized version in the
                # privileged channel
                topic = f"{self._prefix_privileged}.{object_type}"
                dict_ = self.value_sanitizer(object_type, object_.to_dict())
                logger.debug("topic: %s, key: %s, value: %s", topic, key, dict_)
                self.send(topic, key=key, value=dict_)
                object_ = anon_object_

        topic = f"{self._prefix}.{object_type}"
        dict_ = self.value_sanitizer(object_type, object_.to_dict())
        logger.debug("topic: %s, key: %s, value: %s", topic, key, dict_)
        self.send(topic, key=key, value=dict_)

    def write_addition(self, object_type: str, object_: ValueProtocol) -> None:
        """Write a single object to the journal"""
        self._write_addition(object_type, object_)
        if self.auto_flush:
            self.flush()

    def write_additions(
        self, object_type: str, objects: Iterable[ValueProtocol]
    ) -> None:
        """Write a set of objects to the journal"""
        for object_ in objects:
            self._write_addition(object_type, object_)

        if self.auto_flush:
            self.flush()



# Copyright (C) 2019-2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import math
from typing import Dict, List, cast
from unittest.mock import MagicMock

from confluent_kafka import Producer
import pytest

from swh.core.pytest_plugin import FakeSocket
from swh.journal.client import EofBehavior, JournalClient
from swh.journal.serializers import kafka_to_value, key_to_kafka, value_to_kafka
from swh.model.model import Content, Revision
from swh.model.tests.swh_model_data import TEST_OBJECTS

REV = {
    "message": b"something cool",
    "author": {"fullname": b"Peter", "name": None, "email": b"peter@ouiche.lo"},
    "committer": {"fullname": b"Stephen", "name": b"From Outer Space", "email": None},
    "date": {
        "timestamp": {"seconds": 123456789, "microseconds": 123},
        "offset": 120,
        "negative_utc": False,
    },
    "committer_date": {
        "timestamp": {"seconds": 123123456, "microseconds": 0},
        "offset": 0,
        "negative_utc": False,
    },
    "type": "git",
    "directory": (
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x01\x02\x03\x04\x05"
    ),
    "synthetic": False,
    "metadata": None,
    "parents": [],
    "id": b"\x8b\xeb\xd1\x9d\x07\xe2\x1e0\xe2 \x91X\x8d\xbd\x1c\xa8\x86\xdeB\x0c",
}


@pytest.mark.parametrize("legacy_eof", [True, False])
def test_client(
    kafka_prefix: str, kafka_consumer_group: str, kafka_server: str, legacy_eof: bool
):
    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test producer",
            "acks": "all",
        }
    )

    # Fill Kafka
    producer.produce(
        topic=kafka_prefix + ".revision",
        key=REV["id"],
        value=value_to_kafka(REV),
    )
    producer.flush()

    if legacy_eof:
        with pytest.deprecated_call():
            client = JournalClient(
                brokers=[kafka_server],
                group_id=kafka_consumer_group,
                prefix=kafka_prefix,
                stop_on_eof=True,
            )
    else:
        client = JournalClient(
            brokers=[kafka_server],
            group_id=kafka_consumer_group,
            prefix=kafka_prefix,
            on_eof=EofBehavior.STOP,
        )
    worker_fn = MagicMock()
    client.process(worker_fn)

    worker_fn.assert_called_once_with({"revision": [REV]})


def test_client_statsd(
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: str,
):
    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test producer",
            "acks": "all",
        }
    )

    # Fill Kafka
    producer.produce(
        topic=kafka_prefix + ".revision",
        key=REV["id"],
        value=value_to_kafka(REV),
    )
    producer.flush()

    client = JournalClient(
        brokers=[kafka_server],
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        on_eof=EofBehavior.STOP,
    )
    client.statsd._socket = FakeSocket()
    worker_fn = MagicMock()
    client.process(worker_fn)

    group_tag = f"#group:{kafka_consumer_group}"

    worker_fn.assert_called_once_with({"revision": [REV]})
    assert client.statsd.namespace == "swh_journal_client"
    # check at least initialization of the status gauge is ok
    assert (
        client.statsd.socket.recv()
        == f"swh_journal_client.status:0|g|{group_tag},status:idle"
    )
    assert (
        client.statsd.socket.recv()
        == f"swh_journal_client.status:0|g|{group_tag},status:processing"
    )
    assert (
        client.statsd.socket.recv()
        == f"swh_journal_client.status:0|g|{group_tag},status:waiting"
    )
    # processing the batch with only one message
    assert (
        client.statsd.socket.recv()
        == f"swh_journal_client.status:1|g|{group_tag},status:idle"
    )
    assert (
        client.statsd.socket.recv()
        == f"swh_journal_client.status:0|g|{group_tag},status:idle"
    )
    assert (
        client.statsd.socket.recv()
        == f"swh_journal_client.status:1|g|{group_tag},status:waiting"
    )
    assert (
        client.statsd.socket.recv()
        == f"swh_journal_client.status:0|g|{group_tag},status:waiting"
    )
    assert (
        client.statsd.socket.recv()
        == f"swh_journal_client.status:1|g|{group_tag},status:processing"
    )
    assert (
        client.statsd.socket.recv()
        == f"swh_journal_client.status:0|g|{group_tag},status:processing"
    )
    assert (
        client.statsd.socket.recv()
        == f"swh_journal_client.status:1|g|{group_tag},status:idle"
    )
    # from there, status messages can be mixed with a few other ones...
    assert (
        f"swh_journal_client.handle_message_total:1|c|{group_tag}".encode()
        in client.statsd.socket.payloads
    )
    assert (
        f"swh_journal_client.stop_total:1|c|{group_tag}".encode()
        in client.statsd.socket.payloads
    )
    # and some waiting/idle forth and back may have happened, so only check the
    # last 3 messages resetting the status gauges)
    assert list(client.statsd.socket.payloads)[-3:] == [
        f"swh_journal_client.status:0|g|{group_tag},status:idle".encode(),
        f"swh_journal_client.status:0|g|{group_tag},status:processing".encode(),
        f"swh_journal_client.status:0|g|{group_tag},status:waiting".encode(),
    ]


@pytest.mark.parametrize(
    "count,legacy_eof", [(1, True), (2, True), (1, False), (2, False)]
)
def test_client_stop_after_objects(
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: str,
    count: int,
    legacy_eof: bool,
):
    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test producer",
            "acks": "all",
        }
    )

    # Fill Kafka
    revisions = cast(List[Revision], TEST_OBJECTS["revision"])
    for rev in revisions:
        producer.produce(
            topic=kafka_prefix + ".revision",
            key=rev.id,
            value=value_to_kafka(rev.to_dict()),
        )
    producer.flush()

    if legacy_eof:
        with pytest.deprecated_call():
            client = JournalClient(
                brokers=[kafka_server],
                group_id=kafka_consumer_group,
                prefix=kafka_prefix,
                stop_on_eof=False,
                stop_after_objects=count,
            )
    else:
        client = JournalClient(
            brokers=[kafka_server],
            group_id=kafka_consumer_group,
            prefix=kafka_prefix,
            on_eof=EofBehavior.CONTINUE,
            stop_after_objects=count,
        )

    worker_fn = MagicMock()
    client.process(worker_fn)

    # this code below is not pretty, but needed since we have to deal with
    # dicts (so no set) which can have values that are list vs tuple, and we do
    # not know for sure how many calls of the worker_fn will happen during the
    # consumption of the topic...
    worker_fn.assert_called()
    revs = []  # list of (unique) rev dicts we got from the client
    for call in worker_fn.call_args_list:
        callrevs = call[0][0]["revision"]
        for rev in callrevs:
            assert Revision.from_dict(rev) in revisions
            if rev not in revs:
                revs.append(rev)
    assert len(revs) == count


assert len(TEST_OBJECTS["revision"]) < 10, (
    'test_client_restart_and_stop_after_objects expects TEST_OBJECTS["revision"] '
    "to have less than 10 objects to test exhaustively"
)


@pytest.mark.parametrize(
    "count,string_eof", [(1, True), (2, False), (10, True), (20, False)]
)
def test_client_restart_and_stop_after_objects(
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: str,
    count: int,
    string_eof: bool,
):
    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test producer",
            "acks": "all",
        }
    )

    # Fill Kafka
    revisions = cast(List[Revision], TEST_OBJECTS["revision"])
    for rev in revisions:
        producer.produce(
            topic=kafka_prefix + ".revision",
            key=rev.id,
            value=value_to_kafka(rev.to_dict()),
        )
    producer.flush()

    client = JournalClient(
        brokers=[kafka_server],
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        on_eof="restart" if string_eof else EofBehavior.RESTART,
        stop_after_objects=count,
    )

    worker_fn = MagicMock()
    client.process(worker_fn)

    # this code below is not pretty, but needed since we have to deal with
    # dicts (so no set) which can have values that are list vs tuple, and we do
    # not know for sure how many calls of the worker_fn will happen during the
    # consumption of the topic...
    worker_fn.assert_called()
    revs = []  # list of (possibly duplicated) rev dicts we got from the client
    unique_revs = []  # list of (unique) rev dicts we got from the client
    for call in worker_fn.call_args_list:
        callrevs = call[0][0]["revision"]
        for rev in callrevs:
            assert Revision.from_dict(rev) in revisions
            if rev not in revs:
                unique_revs.append(rev)
            revs.append(rev)
    assert len(revs) == count
    assert len(unique_revs) == min(len(revisions), count)

    # Each revision should be seen approximately count/len(revisions) times
    rev_ids = [r["id"].hex() for r in revs]  # type: ignore
    for rev in revisions:
        assert (
            math.floor(count / len(revisions))
            <= rev_ids.count(rev.id.hex())
            <= math.ceil(count / len(revisions))
        )

    # Check each run but the last contains all revisions
    for i in range(int(count / len(revisions))):
        assert set(rev_ids[i * len(revisions) : (i + 1) * len(revisions)]) == set(
            rev.id.hex() for rev in revisions
        ), i


@pytest.mark.parametrize("batch_size", [1, 5, 100])
def test_client_batch_size(
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: str,
    batch_size: int,
):
    num_objects = 2 * batch_size + 1
    assert num_objects < 256, "Too many objects, generation will fail"

    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test producer",
            "acks": "all",
        }
    )

    contents = [Content.from_data(bytes([i])) for i in range(num_objects)]

    # Fill Kafka
    for content in contents:
        producer.produce(
            topic=kafka_prefix + ".content",
            key=key_to_kafka(content.sha1),
            value=value_to_kafka(content.to_dict()),
        )

    producer.flush()

    client = JournalClient(
        brokers=[kafka_server],
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        stop_on_eof=True,
        batch_size=batch_size,
    )

    collected_output: List[Dict] = []

    def worker_fn(objects):
        received = objects["content"]
        assert len(received) <= batch_size
        collected_output.extend(received)

    client.process(worker_fn)

    expected_output = [content.to_dict() for content in contents]
    assert len(collected_output) == len(expected_output)

    for output in collected_output:
        assert output in expected_output


@pytest.fixture()
def kafka_producer(kafka_prefix: str, kafka_server_base: str):
    producer = Producer(
        {
            "bootstrap.servers": kafka_server_base,
            "client.id": "test producer",
            "acks": "all",
        }
    )

    # Fill Kafka
    producer.produce(
        topic=kafka_prefix + ".something",
        key=key_to_kafka(b"key1"),
        value=value_to_kafka("value1"),
    )
    producer.produce(
        topic=kafka_prefix + ".else",
        key=key_to_kafka(b"key1"),
        value=value_to_kafka("value2"),
    )
    producer.flush()
    return producer


def test_client_subscribe_all(
    kafka_producer: Producer, kafka_prefix: str, kafka_server_base: str
):
    client = JournalClient(
        brokers=[kafka_server_base],
        group_id="whatever",
        prefix=kafka_prefix,
        stop_on_eof=True,
    )
    assert set(client.subscription) == {
        f"{kafka_prefix}.something",
        f"{kafka_prefix}.else",
    }

    worker_fn = MagicMock()
    client.process(worker_fn)
    worker_fn.assert_called_once_with(
        {
            "something": ["value1"],
            "else": ["value2"],
        }
    )


def test_client_subscribe_one_topic(
    kafka_producer: Producer, kafka_prefix: str, kafka_server_base: str
):
    client = JournalClient(
        brokers=[kafka_server_base],
        group_id="whatever",
        prefix=kafka_prefix,
        stop_on_eof=True,
        object_types=["else"],
    )
    assert client.subscription == [f"{kafka_prefix}.else"]

    worker_fn = MagicMock()
    client.process(worker_fn)
    worker_fn.assert_called_once_with({"else": ["value2"]})


def test_client_subscribe_absent_topic(
    kafka_producer: Producer, kafka_prefix: str, kafka_server_base: str
):
    with pytest.raises(ValueError):
        JournalClient(
            brokers=[kafka_server_base],
            group_id="whatever",
            prefix=kafka_prefix,
            stop_on_eof=True,
            object_types=["really"],
        )


def test_client_subscribe_absent_prefix(
    kafka_producer: Producer, kafka_prefix: str, kafka_server_base: str
):
    with pytest.raises(ValueError):
        JournalClient(
            brokers=[kafka_server_base],
            group_id="whatever",
            prefix="wrong.prefix",
            stop_on_eof=True,
        )
    with pytest.raises(ValueError):
        JournalClient(
            brokers=[kafka_server_base],
            group_id="whatever",
            prefix="wrong.prefix",
            stop_on_eof=True,
            object_types=["else"],
        )


def test_client_subscriptions_with_anonymized_topics(
    kafka_prefix: str, kafka_consumer_group: str, kafka_server_base: str
):
    producer = Producer(
        {
            "bootstrap.servers": kafka_server_base,
            "client.id": "test producer",
            "acks": "all",
        }
    )

    # Fill Kafka with revision object on both the regular prefix (normally for
    # anonymized objects in this case) and privileged one
    producer.produce(
        topic=kafka_prefix + ".revision",
        key=REV["id"],
        value=value_to_kafka(REV),
    )
    producer.produce(
        topic=kafka_prefix + "_privileged.revision",
        key=REV["id"],
        value=value_to_kafka(REV),
    )
    producer.flush()

    # without privileged "channels" activated on the client side
    client = JournalClient(
        brokers=[kafka_server_base],
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        stop_on_eof=True,
        privileged=False,
    )
    # we only subscribed to "standard" topics
    assert client.subscription == [kafka_prefix + ".revision"]

    # with privileged "channels" activated on the client side
    client = JournalClient(
        brokers=[kafka_server_base],
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        privileged=True,
    )
    # we only subscribed to "privileged" topics
    assert client.subscription == [kafka_prefix + "_privileged.revision"]


def test_client_subscriptions_without_anonymized_topics(
    kafka_prefix: str, kafka_consumer_group: str, kafka_server_base: str
):
    producer = Producer(
        {
            "bootstrap.servers": kafka_server_base,
            "client.id": "test producer",
            "acks": "all",
        }
    )

    # Fill Kafka with revision objects only on the standard prefix
    producer.produce(
        topic=kafka_prefix + ".revision",
        key=REV["id"],
        value=value_to_kafka(REV),
    )
    producer.flush()

    # without privileged channel activated on the client side
    client = JournalClient(
        brokers=[kafka_server_base],
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        stop_on_eof=True,
        privileged=False,
    )
    # we only subscribed to the standard prefix
    assert client.subscription == [kafka_prefix + ".revision"]

    # with privileged channel activated on the client side
    client = JournalClient(
        brokers=[kafka_server_base],
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        stop_on_eof=True,
        privileged=True,
    )
    # we also only subscribed to the standard prefix, since there is no priviled prefix
    # on the kafka broker
    assert client.subscription == [kafka_prefix + ".revision"]


def test_client_with_deserializer(
    kafka_prefix: str, kafka_consumer_group: str, kafka_server: str
):
    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test producer",
            "acks": "all",
        }
    )

    # Fill Kafka
    revisions = cast(List[Revision], TEST_OBJECTS["revision"])
    for rev in revisions:
        producer.produce(
            topic=kafka_prefix + ".revision",
            key=rev.id,
            value=value_to_kafka(rev.to_dict()),
        )
    producer.flush()

    def custom_deserializer(object_type, msg):
        assert object_type == "revision"
        obj = kafka_to_value(msg)
        # filter the first revision
        if obj["id"] == revisions[0].id:
            return None
        return Revision.from_dict(obj)

    client = JournalClient(
        brokers=[kafka_server],
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        stop_on_eof=True,
        value_deserializer=custom_deserializer,
    )
    worker_fn = MagicMock()
    client.process(worker_fn)

    # a commit seems to be needed to prevent some race condition situation
    # where the worker_fn has not yet been called at this point (not sure how)
    client.consumer.commit()

    # Check the first revision has not been passed to worker_fn
    processed_revisions = set(worker_fn.call_args[0][0]["revision"])
    assert revisions[0] not in processed_revisions
    assert all(rev in processed_revisions for rev in revisions[1:])



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict, Iterable, Optional, TypeVar

from typing_extensions import Protocol, runtime_checkable

from swh.model.model import KeyType

TSelf = TypeVar("TSelf")


class ValueProtocol(Protocol):
    def anonymize(self: TSelf) -> Optional[TSelf]:
        ...

    def unique_key(self) -> KeyType:
        ...

    def to_dict(self) -> Dict[str, Any]:
        ...


@runtime_checkable
class JournalWriterInterface(Protocol):
    def write_addition(self, object_type: str, object_: ValueProtocol) -> None:
        """Add a SWH object of type object_type in the journal."""
        ...

    def write_additions(
        self, object_type: str, objects: Iterable[ValueProtocol]
    ) -> None:
        """Add a list of SWH objects of type object_type in the journal."""
        ...

    def flush(self) -> None:
        """Flush the pending object additions in the backend, if any."""
        ...



# Copyright (C) 2016-2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from enum import Enum
from typing import Any, BinaryIO, Union

import msgpack

from swh.model.model import KeyType


class MsgpackExtTypeCodes(Enum):
    LONG_INT = 1
    LONG_NEG_INT = 2


# this as been copied from swh.core.api.serializer
# TODO refactor swh.core to make this function available
def _msgpack_encode_longint(value):
    # needed because msgpack will not handle long integers with more than 64 bits
    # which we unfortunately happen to have to deal with from time to time
    if value > 0:
        code = MsgpackExtTypeCodes.LONG_INT.value
    else:
        code = MsgpackExtTypeCodes.LONG_NEG_INT.value
        value = -value
    length, rem = divmod(value.bit_length(), 8)
    if rem:
        length += 1
    return msgpack.ExtType(code, int.to_bytes(value, length, "big"))


def msgpack_ext_encode_types(obj):
    if isinstance(obj, int):
        return _msgpack_encode_longint(obj)
    return obj


def msgpack_ext_hook(code, data):
    if code == MsgpackExtTypeCodes.LONG_INT.value:
        return int.from_bytes(data, "big")
    if code == MsgpackExtTypeCodes.LONG_NEG_INT.value:
        return -int.from_bytes(data, "big")

    raise ValueError("Unknown msgpack extended code %s" % code)


# for BW compat
def decode_types_bw(obj):
    if set(obj.keys()) == {b"d", b"swhtype"} and obj[b"swhtype"] == "datetime":
        return datetime.datetime.fromisoformat(obj[b"d"])
    return obj


def stringify_key_item(k: str, v: Union[str, bytes]) -> str:
    """Turn the item of a dict key into a string"""
    if isinstance(v, str):
        return v
    if k == "url":
        return v.decode("utf-8")
    return v.hex()


def pprint_key(key: KeyType) -> str:
    """Pretty-print a kafka key"""

    if isinstance(key, dict):
        return "{%s}" % ", ".join(
            f"{k}: {stringify_key_item(k, v)}" for k, v in key.items()
        )
    elif isinstance(key, bytes):
        return key.hex()
    else:
        return key


def key_to_kafka(key: KeyType) -> bytes:
    """Serialize a key, possibly a dict, in a predictable way"""
    p = msgpack.Packer(use_bin_type=True)
    if isinstance(key, dict):
        return p.pack_map_pairs(sorted(key.items()))
    else:
        return p.pack(key)


def kafka_to_key(kafka_key: bytes) -> KeyType:
    """Deserialize a key"""
    return msgpack.loads(kafka_key, raw=False)


def value_to_kafka(value: Any) -> bytes:
    """Serialize some data for storage in kafka"""
    return msgpack.packb(
        value,
        use_bin_type=True,
        datetime=True,  # encode datetime as msgpack.Timestamp
        default=msgpack_ext_encode_types,
    )


def kafka_to_value(kafka_value: bytes) -> Any:
    """Deserialize some data stored in kafka"""
    return msgpack.unpackb(
        kafka_value,
        raw=False,
        object_hook=decode_types_bw,
        ext_hook=msgpack_ext_hook,
        strict_map_key=False,
        timestamp=3,  # convert Timestamp in datetime objects (tz UTC)
    )


def kafka_stream_to_value(file_like: BinaryIO) -> msgpack.Unpacker:
    """Return a deserializer for data stored in kafka"""
    return msgpack.Unpacker(
        file_like,
        raw=False,
        object_hook=decode_types_bw,
        ext_hook=msgpack_ext_hook,
        strict_map_key=False,
        use_list=False,
        timestamp=3,  # convert Timestamp in datetime objects (tz UTC)
    )
